/* Intercepts packet from given netfilter hook, and outputs to stdout:
   Useful for testing with rcv_ip. */
#include <sys/types.h>
#include <limits.h>
#include <net/if.h>

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <sys/time.h>
#include <sys/resource.h>

// Antigone Include Files
#include <AGroup.h>
#include <AtkThread.h>
#include <AtkTimer.h>
#include <AtkThreadLock.h>
#include <Environment.h>
#include <ASelect.h>
//mapping.h is a local include file
#include <mapping.h>
#include "spoofit1.h"

class Listmap sockhandles;
AuditLog *alog;
extern int errno;
char *local_vip = NULL;    // Virtual IP address for the local node

//#define LOGGING
#define ANTIGONE
//#define DEBUG
#ifdef LOGGING
#define LOG(fmt) fprintf(stdout, fmt); \
                 fprintf(stdout,"\n"); \
                 fflush(stdout)
#define LOG1(fmt, arg1) fprintf(stdout, fmt, arg1); \
                        fprintf(stdout,"\n"); \
                        fflush(stdout)

#define LOG2(fmt, arg1, arg2) fprintf(stdout, fmt, arg1, arg2); \
                              fprintf(stdout,"\n"); \
                              fflush(stdout)
#define LOG3(fmt, arg1, arg2, arg3 ) fprintf(stdout, fmt, arg1, arg2, arg3); \
                              fprintf(stdout,"\n"); \
                              fflush(stdout)
#endif
#ifndef LOGGING
#define LOG(fmt)
#define LOG1(fmt, arg1)
#define LOG2(fmt, arg1, arg2)
#define LOG3(fmt, arg1, arg2, arg3 )
#endif

extern "C" {
#include <netinet/ip.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <linux/netfilter_ipv4.h>
#include <linux/netfilter.h>
#include <netinet/ip_icmp.h>
#include <netdb.h>
#include <libipq.h>

#define BUFSIZE 20480
#define MAXMSG 30000
#define SPORT_DEF 5001
#define SADDR_DEF "192.168.27.5"
#define MODE ATransport::AT_SYMMETRIC
#define DEFAULT_POLICY "loc.apd"

#define IPHDR_SIZE (sizeof(struct iphdr))
#define offsetUDP  IPHDR_SIZE
#define offsetTCP  IPHDR_SIZE
#define LOCAL_ADD "192.168.27.5"
#define IP_ADDR "192.168.27.5"
#define WATER "192.168.27.5"

#define TCP_PROTO 6
#define UDP_PROTO 17
#define ICMP_PROTO 1
// possible input sources
#define READ_GROUP (0)
#define READ_SKT (1)

int errorno;
  
static void die(struct ipq_handle *h);

unsigned short in_cksum1(unsigned short *addr,int len)
{
register int nleft = len;
register unsigned short *w = addr;
register int sum = 0;
unsigned short answer = 0;
        
while (nleft > 1)
        { 
        sum += *w++;
        nleft -= 2;
        }
if (nleft == 1)
        {
        *(u_char *)(&answer) = *(u_char *)w ;
        sum += answer;
        }
sum = (sum >> 16) + (sum & 0xffff);
sum += (sum >> 16);
answer = ~sum;
return(answer);
}

u_short cksum(u_short *buf, int nwords) 
{
   unsigned long sum;
   printf("number of words = %d\n",nwords);

   for (sum = 0; nwords > 0; nwords--)
      sum += *buf++;
   sum = (sum >> 16) + (sum & 0xffff);
   sum += (sum >> 16);
   return ~sum ;
}

static int
string_to_number(const char *s, int min, int max)
{
	int number;
	char *end;

	/* Handle hex, octal, etc. */
	number = (int)strtol(s, &end, 0);
	if (*end == '\0' && end != s) {
		/* we parsed a number, let's see if we want this */
		if (min <= number && number <= max)
			return number;
	}
	return -1;
}

static struct in_addr *
dotted_to_addr(const char *dotted)
{
	static struct in_addr addr;
	unsigned char *addrp;
	char *p, *q;
	int onebyte, i;
	char buf[20];

	/* copy dotted string, because we need to modify it */
	strncpy(buf, dotted, sizeof(buf) - 1);
	addrp = (unsigned char *) &(addr.s_addr);

	p = buf;
	for (i = 0; i < 3; i++) {
		if ((q = strchr(p, '.')) == NULL)
			return (struct in_addr *) NULL;
		else {
			*q = '\0';
			if ((onebyte = string_to_number(p, 0, 255)) == -1)
				return (struct in_addr *) NULL;
			else
				addrp[i] = (unsigned char) onebyte;
		}
		p = q + 1;
	}

	/* we've checked 3 bytes, now we check the last one */
	if ((onebyte = string_to_number(p, 0, 255)) == -1)
		return (struct in_addr *) NULL;
	else
		addrp[3] = (unsigned char) onebyte;

	return &addr;
}

/* this function parses the packet for the destination port and also changes the
   source ip to the ip of the gateway router */

int parse_pkt(char *pkt)
{
     int i,j,sd;
     int dport,sport,pkt_len;
     struct iphdr *sip_hdr;
     struct tcphdr *stcp_hdr;
     struct udphdr *sudp_hdr;
     struct in_addr addr,addr1,addr2;
     unsigned long int addr3,addr4;
     char *pseudo_pkt;
     struct pseudo_IP_header *sp_help_pseudo;
     char local_add[20]=LOCAL_ADD;

     sip_hdr = (struct iphdr *)pkt;
     LOG(" ##### recv pkt logs (parsing) #####\n\n"); 
     memcpy((char *)&addr,(const void *)&(sip_hdr->daddr),
            sizeof(sip_hdr->daddr));
     LOG1("dest ip = %s\n",inet_ntoa(addr));
     LOG1("packet len = %d\n",htons(sip_hdr->tot_len));

     if ((local_vip == NULL)||
         ((strncmp(inet_ntoa(addr),local_vip,strlen(local_vip)) != 0) 
          && (strncmp(inet_ntoa(addr),"224.27",6) != 0)))
     {
          printf("packet not destined for us\n");
          return(-1);
     }

     if (strncmp(inet_ntoa(addr),"224.27",6) != 0)
          sip_hdr->daddr = inet_addr((const char *)&local_add);
     
#ifdef LOGGING
     memcpy((char *)&addr,(const void *)&(sip_hdr->saddr),
            sizeof(sip_hdr->saddr));
     LOG1("source ip = %s\n",inet_ntoa(addr));
     //sip_hdr->saddr = inet_addr((const char *)"10.10.27.8");
     //memcpy((char *)&(sip_hdr->daddr),(char *)&local_add,sizeof(sip_hdr->daddr));

     LOG("reached here 34");
     pkt_len = htons(sip_hdr->tot_len);
     pseudo_pkt = (char *)malloc(pkt_len-8);
     sp_help_pseudo = (struct pseudo_IP_header *)pseudo_pkt;

     if (pseudo_pkt == NULL) exit(0);

     bzero(pseudo_pkt,pkt_len-8);
     memcpy(pseudo_pkt + 12, pkt + IPHDR_SIZE, pkt_len - IPHDR_SIZE);

     LOG("reached here 43");
     memcpy((char *)&addr1,(const void *)&(sip_hdr->saddr),
            sizeof(sip_hdr->saddr));
     memcpy((char *)&addr2,(const void *)&(sip_hdr->daddr),
            sizeof(sip_hdr->daddr));
     memcpy((char *)&addr3,(const void *)&(sip_hdr->saddr),
            sizeof(sip_hdr->saddr));
     memcpy((char *)&addr4,(const void *)&(sip_hdr->daddr),
            sizeof(sip_hdr->daddr));

     sp_help_pseudo->source = addr3;
     sp_help_pseudo->destination =  addr4;
     sp_help_pseudo->zero_byte = 0;
#endif     
     switch (sip_hdr->protocol) {
     case TCP_PROTO:
          short offset;
          stcp_hdr = (struct tcphdr *)(pkt + offsetTCP);
          // checking the total header length
          dport = ntohs(stcp_hdr->dest);
#ifdef LOGGING
          memcpy((char *)&offset,stcp_hdr+12, sizeof(offset));
          offset = htons(offset);
          offset = offset >> 12;
          LOG2("offset = %d, dport = %d\n",offset,dport);

          bzero(pseudo_pkt+28,2); // zeroing out the earlier checksum
          sp_help_pseudo->TCP_UDP_len = 
               htons(pkt_len - IPHDR_SIZE);   //data_len + tcp_len
          sp_help_pseudo->protocol = 6;
          //stcp_hdr->check = in_cksum((u_short  *)pseudo_pkt, pkt_len-8);
          LOG("reached here4\n");
#endif
          break;

     case UDP_PROTO:
          sudp_hdr = (struct udphdr *)(pkt + offsetUDP);
          dport = ntohs(sudp_hdr->dest);
/*
          bzero(pseudo_pkt+18,2); // zeroing out the earlier checksum
          sp_help_pseudo->TCP_UDP_len = 
          htons(pkt_len - IPHDR_SIZE); //data_len + udp_len
          sp_help_pseudo->protocol = 17;
          //sudp_hdr->check = in_cksum((u_short  *)pseudo_pkt, pkt_len-8);*/
          break;

     case ICMP_PROTO:
          //sicmp_hdr = (struct icmphdr *)(pkt + offsetICMP);
          //dport = ntohs(sicmp_hdr->dest);
          dport = 1000;
          break;
          
     default:
          dport=0;
     }

#ifdef LOGGING
     LOG("reached here5\n");
     //bzero((void *)&(sip_hdr->check),sizeof(short));
     //sip_hdr->check = in_cksum((u_short  *)pkt, sizeof(struct iphdr));
     //memcpy((char *)&addr,(const void *)&(sip_hdr->daddr),
     //       sizeof(sip_hdr->daddr));
     LOG1("new dest ip = %s\n",inet_ntoa(addr));
     LOG(" ######  recv logs done \n\n");
     delete pseudo_pkt;
#endif
     /*sd = open_sending();
       transmit_TCP(sd,(char *)(pkt+28),0,0,(pkt_len-28),
       addr3, 32435,
       addr4,dport,0,0,2);*/
     // transmit_UDP(sd,(char *)(pkt+28),0,(pkt_len-28),(unsigned long int)addr1,
     //           ntohs(sudp_hdr->source),(unsigned long int)addr2,dport);
     // if using transmit_UDP, uncomment the next line also
     // dport = -1;
     return (dport);

}

//////////////////////////////////////////////////////////////////////////////////
//
// Procedure    : comm_s
// Description  : This is the thread procedure spawned by the main process. It 
//                takes care of all the communication with the Antigone server
//                and other members of the group.
//
// Inputs       : sockfd - it is the pointer to the socket handle

// Outputs      : None

void *comm_s(int *sockfd)
{
     String snder;
     int n;
     //char *dat;
     char dat[MAXMSG];
     unsigned char *dat1;
     //char data1[MAXDATA];
     //char type[4];
     int sfd = *sockfd,i=0,local_skt, dlen;
     unsigned long idx = 1, rindx;
     char data[MAXMSG];
     unsigned long speriod = 1;
     struct sockaddr_in local_addr;
     struct timeval tv={0,100};
     AGroup *grp;

     LOGMSG0( alog, PE_DEFAULT_ALEVEL, "comm thread started");
     printf("comm thread started\n");
     // getting the corresponding socket info 
     SockObject *sk;
     if ((sk = sockhandles.mapfd(sfd)) == NULL) {
          errorno = ERR_GROUP;
          LOGMSG0(alog, PE_DEFAULT_ALEVEL,"comm_s: sockhandle error");
          return NULL;
     }
     grp = sk->grp;

     AtkTimer read_timer(speriod);

     // filling up the local address
     local_addr.sin_family = AF_INET;
     local_addr.sin_port=htons(5001);
     local_addr.sin_addr.s_addr = inet_addr(LOCAL_ADD);
     
     if ((local_skt = socket(AF_INET,SOCK_RAW, 255)) < 0) {
          printf("socket creation error\n");
          exit(0);
     }

     //usleep(1000000);
     if (sk->grp->doGroup() != TRUE)
          printf("dogroup error\n");

     //local_vip = (char *)((sk->grp->LocalID()).Virt_IP());
     /*
       while ((local_vip = (char *)((sk->grp->LocalID()).Virt_IP())) == NULL) {
       printf("local IP address = %s\n",(char *)((sk->grp->LocalID()).Virt_IP()));
       usleep(100000);
       }
     printf("local IP address = %s\n",(char *)((sk->grp->LocalID()).Virt_IP()));
     */

     ASelect aselect (alog);
     aselect.Add (sk->grp, (void*)READ_GROUP);
     aselect.Add (sk->send_sockpair[0], (void*)READ_SKT);

     int inloop=0;
     //while ( grp->State() == AGroup::AG_INOPERATION ){
     while (1) {
          Buffer *buf = NULL;

          //pthread_mutex_lock(&ft);
          //LOGMSG0( alog, PE_DEFAULT_ALEVEL, "before reading a message");
          /*
            if (local_vip == NULL) {
            local_vip = (char *)((grp->LocalID()).Virt_IP());
            if (local_vip != NULL) {
            int len = strlen(local_vip)+1;
            local_vip = (char *)malloc(len);
            memset(local_vip,'\0',len);
            printf("local_vip filled with len = %d\n",len);
            memcpy(local_vip,(char *)((sk->grp->LocalID()).Virt_IP()),len-1);
            }
            }*/

          LOG1("in loop %d",inloop++);
          //read_timer.reset();
          ASelectInput *in = aselect.Wait();
          // checking for virtual ip again
          if (local_vip == NULL) {
            local_vip = (char *)((grp->LocalID()).Virt_IP());
            if (local_vip != NULL) {
              int len = strlen(local_vip)+1;
              local_vip = (char *)malloc(len);
              memset(local_vip,'\0',len);
              printf("local_vip filled with len = %d\n",len);
              memcpy(local_vip,(char *)((grp->LocalID()).Virt_IP()),len-1);
            }
          }

          switch ((int)in->getData())
          {
          case READ_GROUP:
               LOG("Select: read available from group");
               if ( (grp->readMessage( &buf , &tv) == TRUE) 
                    && (buf != NULL) )
               {
                    //LOGMSG0( alog, PE_DEFAULT_ALEVEL, "just read a message");  
                    //pthread_mutex_unlock(&ft);
                    // Read message, log it
                    //LOGMSG0( alog, PE_DEFAULT_ALEVEL, "decoding the message");

                    dlen = buf->Length();
                    dat1 = buf->Buf();
                    //LOGMSG1(alog, PE_DEFAULT_ALEVEL,"recvd pkt length = %d",dlen);
                    //printf("pkt recvd , len = %d\n",dlen);
                    //fflush(stdout);
                    //LOGMSG1(alog, PE_DEFAULT_ALEVEL,"message length = %d",dlen);
                    //dat = (char *)malloc(dlen);
                    //memset(&dat,0x0,MAXMSG);
                    
                    //if (dlen>0) buf->getFixedLength((unsigned char *)&dat,dlen);
                    //LOGMSG1(alog, PE_DEFAULT_ALEVEL,"message = %s",dat);

                    //sending the received data to the local socket
                    if (dlen>0) {
                         int dport = parse_pkt((char *)dat1);
                         int nbytes;
                         local_addr.sin_port=htons(dport);
                         if (dport >0){
                              nbytes = sendto(local_skt, (char *)dat1, dlen, 0, 
                                              (struct sockaddr *)&local_addr, 
                                              sizeof(struct sockaddr));
#ifdef DEBUG                              
                              printf("message len on RAW skt = %d\n",dlen);
                              if (nbytes != dlen) {
                                   printf("errno = %d\n",errno);
                                   perror("comm");
                                //printf("socket error, expected %d, sent %d\n"
                                //       ,dlen,nbytes);
                              }
#endif
                         }
                         delete buf;
                    }
               }
               break;
          // sending 
          //LOGMSG0(alog, PE_DEFAULT_ALEVEL, "about to send a message");
          case READ_SKT:
               LOG("Select: read available from socket for sending");
               if (local_vip == NULL) {
                    printf("but local_vip == NULL\n");
                    break;
               }
               //memset((char *)&data,'\0',MAXMSG);
               n = read(sk->send_sockpair[0], &data, MAXMSG);
               if (n > 0) {
                    // writing the virtual source address
                    struct iphdr *sip_hdr = (struct iphdr *)&data;
                    LOG1("local virtual ip address = %s\n",local_vip);
                    sip_hdr->saddr = inet_addr((const char *)local_vip);
                    
                    //Buffer *buf = new Buffer((char *)data, n);
                    //buf->putFixedLength((unsigned char *) data,n);
                    grp->sendMessage( (char *)data, n );
                    //LOG("comm_s: Sent a APP packet");
                    //printf("comm_s: Sent a APP packet\n");
               }
               break;
               //printf("comm_s: loop completed\n");
               //LOGMSG0(alog, PE_DEFAULT_ALEVEL,"comm_s: loop completed");
               
               //pthread_mutex_unlock(&ft);
               //usleep(1);
          }
     }
     LOGMSG0(alog, PE_DEFAULT_ALEVEL,"group not connected now");
     printf("group not connected now\n");
}

//////////////////////////////////////////////////////////////////////////////////
//
// Function     : socket_s
// Description  : socket constructor
//
// Inputs       : family - domain of the socket (ex - AF_UNIX, ..)
//                type - type of the desired socket (STREAM, DATAGRAM, ..)
//                protocol - protocol to be used by the socket
// Outputs      : socket handle for future references

int socket_s(int family, int type, int protocol)
{
     char *alogs;
     if (family!=AF_INET) return(-1);

     // currently Socket_s library only supports multicast 
     if (type!=SOCK_DGRAM) {
          errorno = SOCK_ERR;
          return(-1);
     }

     // valid protocol checking
     if ((protocol!=IPPROTO_IP)&&(protocol!=IPPROTO_TCP)&&(protocol!=IPPROTO_UDP))
          return(-1);

     //AGroup *group = NULL;
     int fd = sockhandles.Insert(NULL,type,DEFAULT_POLICY);

     //  log creation for logging all the debug information
     alogs = getenv("NAME");
     alog = new AuditLog( alogs, FALSE, FALSE , TRUE);
     alog->setAuditLevel( PE_DEFAULT_ALEVEL );

     //saddr = (unsigned long)inet_addr("192.168.27.1");
     return(fd);
}


//////////////////////////////////////////////////////////////////////////////////
//
// Function     : join_mult
// Description  : joins a socket to a multicast group (using Antigone). Internal
//                function.
//
// Inputs       : sockfd - socket handle
//                mult_addr - multicast address data structure for multicast group
//
// Outputs      : procedure status

int join_mult(int  sockfd, struct ip_mreq *mult_addr)
{
     SockObject *sk;
     char *saddr = getenv("SADDR");
     char *port_str = getenv("SPORT");
     int sport;
     static int sfd=sockfd;
     int cpd;
     ATransport *trans;

     if (sockfd < 0) {
          errorno = INVALID_SOCKET;
          return (-1);
     }

     if ((sk = sockhandles.mapfd(sockfd)) == NULL)
          return(-1);

     String *lpol=new String(sk->pol,-1);
     LOGMSG1(alog,PE_DEFAULT_ALEVEL,"The policy file is %s",sk->pol);
     
     ATransport::AtMode mode = MODE;
     char *modenv = getenv("MODE");

     if (strcmp(modenv,"ASYM") == 0) {
          mode = ATransport::AT_ASYMMETRIC;
          LOGMSG0(alog, PE_DEFAULT_ALEVEL, "MODE is ASYMMETRIC");
     }
     if (strcmp(modenv,"PTOP") == 0) {
          mode = ATransport::AT_PTTOPT;
          LOGMSG0(alog, PE_DEFAULT_ALEVEL, "MODE is PTTOPT");
     }
     if (mode == MODE) 
          LOGMSG0(alog, PE_DEFAULT_ALEVEL, "MODE is SYMMETRIC");

     IPAddress *ipsaddr = NULL;

     if (saddr == NULL) // no environment definition
          saddr = SADDR_DEF;

     if (mode != ATransport::AT_SYMMETRIC)
          ipsaddr = IPAddress::IPAddressFactory(saddr , 5001);

     int portno = sk->port;
     if (portno < 1)
          portno = 5000; // default port allocation

     if (port_str != NULL) 
          sport = atoi(port_str);
     else // no environment definition
          sport = SPORT_DEF;

     // create a symmetric transport object

     IPAddress *lipaddr;
     ATransport::SetLoopback( FALSE );
     
     LOGMSG0(alog, PE_DEFAULT_ALEVEL,"setting up the IPAddress structures");

     if (mode == ATransport::AT_PTTOPT)
          lipaddr = IPAddress::IPAddressFactory( saddr, sport);
     else
          lipaddr = IPAddress::IPAddressFactory( mult_addr->imr_multiaddr.s_addr , portno );

     if ( lipaddr == NULL )
     {
          // Error out
          LOGMSG0(alog,PE_DEFAULT_ALEVEL,"join_mult: IPAddress Error"); 
          errorno = UNKNOWN_ERROR;
          return( -1 );
     }
  
     if ( mode == ATransport::AT_PTTOPT )
          // unicast address is used as (virtual) mcast address on P2P
          trans = new ATransport( (IPAddress *)NULL, lipaddr, mode );
     else
          trans = new ATransport( lipaddr, ipsaddr, mode );
  
     AGroup *grp = new AGroup(trans, *lpol, NULL, NULL, alog);
     trans->Log() = alog;
     //grp->LocalPolicy()->logPolicy( PE_DEFAULT_ALEVEL, alog );

     // setting up the socket handle mapping
     if (sockhandles.changegp(sockfd,grp) < 0) {
          errorno = ERR_GROUP;
          return(-1);
     }

     LOGMSG0(alog, PE_DEFAULT_ALEVEL, "connecting to the group");
     if ( grp->Connect() != TRUE )
     {
          // Error out
          LOGMSG0( alog, FATAL_ERROR, "Error connecting/joining group." );
          errorno = GRP_JOIN;
          return(-1);
     }
     //SockObject *sk;
     if ((sk = sockhandles.mapfd(sockfd)) == NULL) {
          errorno = ERR_GROUP;
          return (-1);
     }

     //creating a socketpair for communication with thread "comm_s"
     if (socketpair(AF_UNIX, SOCK_DGRAM, 0, sk->recv_sockpair) < 0) {
          errorno = ERR_GROUP;
          return(-1);
     }

     if (socketpair(AF_UNIX, SOCK_DGRAM, 0, sk->send_sockpair) < 0) {
          errorno = ERR_GROUP;
          return(-1);
     }
  
     // Setting reading socket of the pair to be non-blocking
     fcntl(sk->recv_sockpair[0],F_SETFL, O_NONBLOCK);
     fcntl(sk->send_sockpair[0],F_SETFL, O_NONBLOCK);
     /*
       if ((cpd = fork()) < 0) {
       LOGMSG0(alog, PE_DEFAULT_ALEVEL, "(join_mult): fork failed");
       printf("fork failed\n");
       exit(-1);
       }
       if (cpd == 0) {
       comm_s(&sfd);
       LOGMSG0(alog, PE_DEFAULT_ALEVEL,"thread exited");
       printf("child process exited\n");
       return(1);
       }
     */
     AtkThread *th = new AtkThread((AtkThreadFuncPtr) &comm_s, &sfd);
     th->Launch();
     unsigned int in=1; 

}

int main(int argc, const char *argv[])
{
	int rval, user_verdict;
	unsigned int hookmask, count, wait, mangle, timeout, i;
	unsigned char packet[BUFSIZE];
	struct ipq_handle *h;
	u_int32_t srcip;
        int joined=0,sockfd,fromport,destport;
        const char multadd[20] = "224.27.27.26";
        struct sockaddr_in multaddr;
        struct in_addr addr;
        struct ip_mreq madd;
        SockObject *sk;
        struct iphdr *sip_hdr;
        struct tcphdr *stcp_hdr;
        struct udphdr *sudp_hdr;

        setpriority(PRIO_PGRP, 0 , -20);

        multaddr.sin_family = AF_INET;
        multaddr.sin_port = htons(5000);
        multaddr.sin_addr.s_addr = inet_addr(multadd); 

        madd.imr_multiaddr.s_addr = inet_addr(multadd);
        madd.imr_interface.s_addr = htonl(INADDR_ANY);
        
	if (argc < 0 || argc > 7) {
		fprintf(stderr,
			"Usage: %s hook-name verdict timeout count"
			"  [src=ipaddr] [wait] [`mangle']\n",
			argv[0]);
		return 1;
	}

	h = ipq_create_handle(0);
	if (!h) {
             printf("here 1\n");
             die(h);
        }
	
	/* The queue does not need to be explicitly flushed now */
	
	/* Tell the queue to send packets up to size BUFSIZE */
	rval = ipq_set_mode(h, IPQ_COPY_PACKET, BUFSIZE);
	if (rval < 0) {
             printf("here 2\n");
             die(h);
        }
        hookmask=1 << NF_IP_LOCAL_IN;
        /*
          if (strcasecmp(argv[1], "LOCAL_IN") == 0)
          hookmask = 1 << NF_IP_LOCAL_IN;
          else if (strcasecmp(argv[1], "PRE_ROUTING") == 0)
          hookmask = 1 << NF_IP_PRE_ROUTING;
          else if (strcasecmp(argv[1], "FORWARD") == 0)
          hookmask = 1 << NF_IP_FORWARD;
          else if (strcasecmp(argv[1], "POST_ROUTING") == 0)
          hookmask = 1 << NF_IP_POST_ROUTING;
          else if (strcasecmp(argv[1], "LOCAL_OUT") == 0)
          hookmask = 1 << NF_IP_LOCAL_OUT;
          else {
          fprintf(stderr, "Unknown hook `%s'\n", argv[1]);
          return 1;
          }*/

        user_verdict=NF_DROP;
/*
	if (strcasecmp(argv[2], "ACCEPT") == 0)
		user_verdict = NF_ACCEPT;
	else if (strcasecmp(argv[2], "DROP") == 0)
		user_verdict = NF_DROP;
	else {
		fprintf(stderr, "Illegal verdict `%s'\n", argv[2]);
		return 1;
	}
*/
	timeout = 10000; //atoi(argv[3]);
	if (timeout < 1) {
		fprintf(stderr, "Illegal timeout `%s'\n", argv[3]);
		return 1;
	}

	count = 10000; //atoi(argv[4]);
	if (count < 1) {
		fprintf(stderr, "Illegal count `%s'\n", argv[4]);
		return 1;
	}

	if (argc > 5 && strncmp(argv[5], "src=", 4) == 0) {
		const struct in_addr *addr;

		addr = dotted_to_addr(argv[5]+4);
		if (addr == NULL) {
			fprintf(stderr, "Illegal src `%s'\n", argv[5]);
			return 1;
		}
		argc--; argv++;
		srcip = addr->s_addr;
	} else
		srcip = 0;

	if (argc > 5) {
		wait = atoi(argv[5]);
		if (wait < 1) {
			fprintf(stderr, "Illegal wait `%s'\n", argv[5]);
			return 1;
		}
		argc--; argv++;
	} else
		wait = 0;

	if (argc > 5 && strcmp(argv[5], "mangle") == 0) {
		mangle = 1;
		argc--; argv++;
	} else
		mangle = 0;

	if (argc > 5) {
		fprintf(stderr, "Error in args: too many.\n");
		exit(1);
	}

        sockfd = socket_s(AF_INET, SOCK_DGRAM, IPPROTO_IP);
        join_mult(sockfd,&madd);
        joined=1;
        
        if ((sk = sockhandles.mapfd(sockfd)) == NULL) {
             errorno = ERR_GROUP;
             printf("send_s: couldn't find the group");
             LOGMSG0(alog, PE_DEFAULT_ALEVEL,
                     "send_s: couldn't find the group");
             return (-1);
        }
         //SockObject *sk;

	while (1) {

		/* We'll die if we receive this. OK. */
          //alarm(timeout);

	try_again:
             //memset(&packet,0x0,BUFSIZE);
             LOG1("i = %d\n",i);
		rval = ipq_read(h, packet, BUFSIZE, 0);
                LOG("packet recvd");
                //fflush(stdout);
		if (rval < 0) {
                     printf("here 3\n");
                     die(h);
                }
		switch (ipq_message_type(packet)) {
		
		case NLMSG_ERROR:
			fprintf(stderr, "Received error message %d %s\n",
				ipq_get_msgerr(packet),
                                strerror(ipq_get_msgerr(packet)));
			break;//exit(1);

                case IPQM_PACKET: {
                     //unsigned char ehdr[16];
                     ipq_packet_msg_t *m = ipq_get_packet(packet);
                     unsigned char *payload = m->payload;
                     
                     //rval = ipq_set_verdict(h, m->packet_id,
                     //                       user_verdict, 0, NULL);

                     if (rval == 0)
                          goto try_again;
#ifdef LOGGING

                     LOG("\n");

                     /* Bogus ethernet header for rcv_ip */
                     memset(&ehdr, 'E', sizeof(ehdr));
                     //write(STDOUT_FILENO, &ehdr, sizeof(ehdr));
                     //write(STDOUT_FILENO, payload + 28, m->data_len);

                     /* This usually wipes out alarm, but that's OK since
                        the rest shouldn't block. */

                     // doing the antigone stuff
                     // analysing the source

                     // if null ip header, terminate current loop

                     stcp_hdr = (struct tcphdr *)(payload + offsetTCP);
                     fromport = ntohs(stcp_hdr->source);
                     destport = ntohs(stcp_hdr->dest);
                     LOG("    ****  sending pkt logs **** \n");
                     LOG2("source port = %d; dest port = %d\n",
                          fromport,destport);
                     //printf("icmp hdr size = %d\n",sizeof(struct icmphdr));
                     sip_hdr = (struct iphdr *)payload;
                     if (sip_hdr == NULL)
                          break;

                     memcpy((char *)&addr,(const void *)&(sip_hdr->saddr),
                            sizeof(sip_hdr->saddr));
                     LOG1("source ip = %s\n",inet_ntoa(addr));
                     memcpy((char *)&addr,(const void *)&(sip_hdr->daddr),
                            sizeof(sip_hdr->daddr));
                     LOG1("dest ip = %s\n",inet_ntoa(addr));
                     LOG2("packet len = %d, %d\n",htons(sip_hdr->tot_len),m->data_len);
#endif
                     /*
                       if (!joined) {
                       joined=1;
                       }
                       else {
                          
                       if ((sk = sockhandles.mapfd(sockfd)) == NULL) {
                       errorno = ERR_GROUP;
                       printf("send_s: couldn't find the group");
                       LOGMSG0(alog, PE_DEFAULT_ALEVEL,
                       "send_s: couldn't find the group");
                       return (-1);
                       }
                     */
                        
                     //if (strncmp(inet_ntoa(addr),"224.27",6) != 0)
                     //   sip_hdr->saddr = inet_addr(IP_ADDR);
                     /* temporarily setting the destination to be water 
                        sip_hdr->daddr = inet_addr(WATER);*/
                     {
                          //printf("packet len = %d\n",m->data_len);
                          write(sk->send_sockpair[1],
                                (const char *)(payload),
                                m->data_len);
                     } 
                     LOG(" ****** packet sent almost **** \n\n");
                     /* Issue verdict and possibly modified packet*/
                     rval = ipq_set_verdict(h, m->packet_id,
                                            user_verdict, m->data_len, payload);
                     if (rval < 0) {
                          printf("here 4\n");
                          die(h);
                     }
                     if (getenv("VERBOSE") && wait)
                          fprintf(stderr, "intercept: reinjected.\n");
                     i++;	
                     break;
		}
                
		default:
                     fprintf(stderr, "Unknown message!\n");
                     exit(1);
                }
        }
	return 0;
}

static void die(struct ipq_handle *h)
{
     perror("intercept");
     ipq_perror("intercept");
     printf("%s\n",ipq_errstr());
     ipq_destroy_handle(h);
     exit(1);
}
}

